from magiccionary.magic import remove_keys


def test_simple_one_non_existing_key():
    input = {
        "a": 1,
        "b": 2,
        "c": 3,
    }
    expected = {
        "a": 1,
        "b": 2,
        "c": 3,
    }
    actual = remove_keys(input, ["z"])

    assert actual == expected
    assert expected == input


def test_simple_one_non_existing_key_list():
    input = {
        "a": {
            "b": {
                "c": {
                    "d": 1,
                },
            },
        }
    }
    expected = {
        "a": {
            "b": {
                "c": {
                    "d": 1,
                },
            },
        }
    }
    actual = remove_keys(input, [["a", "c", "*", "z"]])

    assert actual == expected
    assert expected == input


def test_mixed_list():
    input = {
        "a": {
            "aa": {
                "one": [{"a": 1}, {"b": 2}],
                "two": 2,
            },
            "bb": {
                "one": 1,
                "two": 2,
            },
        },
        "b": {
            "bb": {
                "one": 1,
                "two": 2,
            },
        },
    }

    expected = {
        "a": {
            "aa": {
                "one": [{}, {"b": 2}],
                "two": 2,
            },
            "bb": {
                "two": 2,
            },
        },
    }
    actual = remove_keys(input, [["*", "*", "one", "[]", "a"], ["a", "bb", "one"], ["b"]])

    assert actual == expected
    assert expected == input


def test_mixed_removables_2():
    input = {
        "a": {
            "aa": {
                "one": [{"a": 1}, {"b": 2}],
                "two": 2,
            },
            "bb": {
                "one": 1,
                "two": 2,
            },
        },
        "b": {
            "bb": {
                "one": 1,
                "two": 2,
            },
        },
    }

    expected = {
        "a": {
            "aa": {
                "one": [{}, {"b": 2}],
                "two": 2,
            },
            "bb": {
                "one": 1,
                "two": 2,
            },
        },
        "b": {
            "bb": {
                "one": 1,
                "two": 2,
            },
        },
    }
    actual = remove_keys(input, [["*", "*", "one", "[]", "a"]])

    assert actual == expected
    assert expected == input


def test_mixed_removables():
    input = {
        "a": {
            "list": [
                {
                    "a": {
                        "one": 1,
                    },
                    "b": {
                        "one": 1,
                        "two": 2,
                    },
                },
                {
                    "a": {
                        "one": 1,
                    },
                    "d": {
                        "one": 1,
                        "two": 2,
                    },
                },
            ],
        }
    }

    expected = {
        "a": {
            "list": [
                {
                    "a": {},
                    "b": {
                        "two": 2,
                    },
                },
                {
                    "a": {},
                    "d": {
                        "two": 2,
                    },
                },
            ],
        }
    }
    actual = remove_keys(input, [["a", "list", "[]", "*", "one"]])

    assert actual == expected
    assert expected == input


def test_arbitrary_dict_keys():
    input = {
        "a": {
            "b1": {
                "hello": "world",
                "c": 1,
            },
            "b2": {
                "hello": "world",
                "c": 2,
            },
            "b3": {
                "hello": "world",
                "d": 2,
            },
        }
    }

    expected = {
        "a": {
            "b1": {"hello": "world"},
            "b2": {"hello": "world"},
            "b3": {"hello": "world", "d": 2},
        }
    }

    actual = remove_keys(input, [["a", "*", "c"]])

    assert actual == expected
    assert expected == input


def test_with_list_level():
    input = {
        "a": [
            {"b": 1},
            {"c": 2},
        ]
    }
    expected = {
        "a": [
            {"b": 1},
            {},
        ]
    }
    actual = remove_keys(input, [["a", "[]", "c"]])

    assert actual == expected
    assert expected == input


def test_simple_one_level():
    input = {
        "a": 1,
        "b": 2,
        "c": 3,
    }
    expected = {
        "a": 1,
        "b": 2,
    }
    actual = remove_keys(input, ["c"])

    assert actual == expected
    assert expected == input


def test_simple_two_level():
    input = {
        "a": {
            "b": 2,
            "c": 3,
        }
    }
    expected = {
        "a": {
            "b": 2,
        }
    }
    actual = remove_keys(input, [["a", "c"]])

    assert actual == expected
    assert expected == input


def test_simple_three_level():
    input = {
        "a": {
            "b": {
                "c": 3,
            }
        },
        "d": 4,
    }
    expected = {
        "a": {"b": {}},
        "d": 4,
    }
    actual = remove_keys(input, [["a", "b", "c"]])

    assert actual == expected
    assert expected == input



def _remove_keys(data, single_key_to_remove):
    first_key = single_key_to_remove[0]
    if first_key == "[]" and isinstance(data, list):
        for item in data:
            _remove_keys(item, single_key_to_remove[1:])
    elif first_key == "*" and isinstance(data, dict):
        for key, value in data.items():
            _remove_keys(value, single_key_to_remove[1:])
    elif len(single_key_to_remove) == 1:
        data.pop(first_key, None)
    elif first_key in data:
        if isinstance(data[first_key], dict):
            _remove_keys(data[first_key], single_key_to_remove[1:])
        elif isinstance(data[first_key], list):
            _remove_keys(data[first_key], single_key_to_remove[1:])


def remove_keys(data, keys_to_remove, separator="."):
    for key in keys_to_remove:
        if isinstance(key, str):
            key = key.split(separator)
        _remove_keys(data, key)
    return data


def keep_keys(data, keys_to_keep, separator="."):
    dict_to_keep = {}
    for key in keys_to_keep:
        if isinstance(key, str):
            key = key.split(separator)
        keep = _keep_keys(data, key)
        nested_update(dict_to_keep, keep)
    return dict_to_keep


def _keep_keys(data, single_key_to_keep):
    dict_to_keep = {}
    first_key = single_key_to_keep[0]

    if first_key == "[]" and isinstance(data, list):
        list_to_keep = []
        for item in data:
            list_to_keep.append(_keep_keys(item, single_key_to_keep[1:]))
        # Return the list as is
        return list_to_keep
    elif first_key == "*" and isinstance(data, dict):
        for key, value in data.items():
            dict_to_keep[key] = _keep_keys(value, single_key_to_keep[1:])
        return dict_to_keep
    elif first_key in data:
        if len(single_key_to_keep) == 1:
            dict_to_keep[first_key] = data[first_key]
        elif isinstance(data[first_key], dict):
            dict_to_keep[first_key] = _keep_keys(data[first_key], single_key_to_keep[1:])
        elif isinstance(data[first_key], list):
            dict_to_keep[first_key] = _keep_keys(data[first_key], single_key_to_keep[1:])
    return dict_to_keep

def remove_empty_keys(data):
    empty_keys = []
    for key, value in data.items():
        if value is None:
            empty_keys.append(key)
        elif isinstance(value, dict):
            remove_empty_keys(value)
            if not value:
                empty_keys.append(key)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    remove_empty_keys(item)
                    if not item:
                        empty_keys.append(key)
                elif isinstance(item, str):
                    if not item:
                        empty_keys.append(key)
    for key in empty_keys:
        del data[key]
    return data



def nested_update(original, update):
    for key, value in update.items():
        if isinstance(value, dict) and key in original and isinstance(original[key], dict):
            nested_update(original[key], value)
        else:
            original[key] = value
    return original



from magiccionary.magic import keep_keys


def test_keep_mixed_list():
    input = {
        "a": {
            "aa": {
                "one": [{"a": 1}, {"b": 2}],
                "two": 2,
            },
            "bb": {
                "one": 1,
                "two": 2,
            },
        },
        "b": {
            "bb": {
                "one": 1,
                "two": 2,
            },
        },
    }

    expected = {
        "a": {
            "aa": {
                "one": [{"a": 1}, {}],
            },
            "bb": {
                "one": 1,
            },
        },
        "b": {
            "bb": {
                "one": 1,
                "two": 2,
            },
        },
    }
    actual = keep_keys(input, [["*", "*", "one", "[]", "a"], ["a", "bb", "one"], ["b"]])

    assert actual == expected


def test_keep_arbitrary_dict_keys():
    input = {
        "a": {
            "b1": {
                "hello": "world",
                "c": 1,
            },
            "b2": {
                "hello": "world",
                "c": 2,
            },
            "b3": {
                "hello": "world",
                "d": 2,
            },
        }
    }

    expected = {
        "a": {
            "b1": {"c": 1},
            "b2": {"c": 2},
            "b3": {},
        }
    }

    actual = keep_keys(input, [["a", "*", "c"]])

    assert actual == expected


def test_keep_with_list_level():
    input = {
        "a": [
            {"b": 1},
            {"c": 2},
        ]
    }
    expected = {
        "a": [
            {},
            {"c": 2},
        ]
    }
    actual = keep_keys(input, [["a", "[]", "c"]])

    assert actual == expected


def test_keep_simple_two_level():
    input = {
        "a": {
            "b": 2,
            "c": 3,
        }
    }
    expected = {
        "a": {
            "b": 2,
        }
    }
    actual = keep_keys(input, [["a", "b"]])

    assert actual == expected


def test_keep_simple_non_existing_key():
    input = {
        "a": 1,
        "b": 2,
        "c": 3,
    }
    input_copy = input.copy()
    expected = {}
    actual = keep_keys(input, ["x", "z"])

    assert actual == expected
    assert input == input_copy


def test_keep_simple():
    input = { "a": 1, "b": 2, "c": 3, }
    expected = { "a": 1, "b": 2, }
    input_copy = input.copy()
    
    actual = keep_keys(input, ["a", "b"])

    assert actual == expected
    assert input == input_copy


